################################################################################
# Module: archetypal.template
# Description:
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import collections
import functools
import math
import sqlite3
import time
from operator import add

import numpy as np
from deprecation import deprecated
from eppy.bunch_subclass import BadEPFieldError
from geomeppy.geom.polygons import Polygon3D
from sigfig import round

from archetypal import __version__, log, settings
from archetypal.template import (
    DomesticHotWaterSetting,
    OpaqueConstruction,
    UmiBase,
    UniqueName,
    VentilationSetting,
    WindowSetting,
    ZoneConditioning,
    ZoneConstructionSet,
    ZoneLoad,
)


class InternalMass(object):
    """Class handles the creation of InternalMass constructions from
    :class:`ZoneDefinition`."""

    @classmethod
    def from_zone(cls, zone, **kwargs):
        """

        Args:
            zone (ZoneDefinition): A ZoneDefinition object.
            **kwargs:

        Returns:
            Construction: The internal mass construction for the zone
            None: if no internal mass defined for zone.
        """
        internal_mass_objs = zone._epbunch.getreferingobjs(
            iddgroups=["Thermal Zones and Surfaces"], fields=["Zone_or_ZoneList_Name"]
        )

        area = 0  # initialize area
        mass_opaque_constructions = []  # collect internal mass objects

        # Looping over possible InternalMass objects
        # This InternalMass object (int_obj) is assigned to self,
        # then create object and append to list. There could be more then
        # one.
        for int_obj in internal_mass_objs:
            if int_obj.key.upper() == "INTERNALMASS":
                mass_opaque_constructions.append(
                    OpaqueConstruction.from_epbunch(int_obj, Category="Internal Mass")
                )
                area += float(int_obj.Surface_Area)

        # If one or more constructions, combine them into one.
        if mass_opaque_constructions:
            # Combine elements and assign the aggregated Surface Area
            zone.InternalMassExposedPerFloorArea = float(area) / zone.area
            return functools.reduce(add, mass_opaque_constructions)
        else:
            # No InternalMass object assigned to this Zone, then return Zone and set
            # floor area to 0
            zone.InternalMassExposedPerFloorArea = 0
            return None

    @classmethod
    def generic_internalmass_from_zone(cls, zone):
        """Assign a generic internal mass with InternalMassExposedPerFloorArea = 0 to zone.

        Also set it to the self.InternalMassConstruction attribute.

        Args:
            zone (ZoneDefinition): A ZoneDefinition object.
        """
        zone.InternalMassConstruction = OpaqueConstruction.generic_internalmass(
            idf=zone.idf
        )
        zone.InternalMassExposedPerFloorArea = 0


class ZoneDefinition(UmiBase):
    """Class containing HVAC settings: Conditioning, Domestic Hot Water, Loads,
    Ventilation, adn Constructions

    .. image:: ../images/template/zoneinfo-zone.png
    """

    def __init__(
        self,
        Name,
        Constructions=None,
        Loads=None,
        Conditioning=None,
        Ventilation=None,
        DomesticHotWater=None,
        DaylightMeshResolution=1,
        DaylightWorkplaneHeight=0.8,
        InternalMassConstruction=None,
        InternalMassExposedPerFloorArea=1.05,
        Windows=None,
        **kwargs,
    ):
        """Initialize :class:`Zone` object.

        Args:
            Name (str): Name of the object. Must be Unique.
            Constructions (ZoneConstructionSet):
            Loads (ZoneLoad): Loads of the zone defined with the lights,
                equipment and occupancy parameters (see :class:`ZoneLoad`)
            Conditioning (ZoneConditioning): Conditioning of the zone defined
                with heating/cooling and mechanical ventilation parameters (see
                :class:`ZoneConditioning`)
            Ventilation (VentilationSetting): Ventilation settings of the zone
                defined with the infiltration rate and natural ventilation
                parameters (see :class:`VentilationSetting`)
            DomesticHotWater (archetypal.template.dhw.DomesticHotWaterSetting):
            DaylightMeshResolution (float):
            DaylightWorkplaneHeight (float):
            InternalMassConstruction (archetypal.OpaqueConstruction):
            InternalMassExposedPerFloorArea:
            Windows (WindowSetting): The WindowSetting object associated with
                this zone.
            **kwargs:
        """
        super(ZoneDefinition, self).__init__(Name, **kwargs)

        self.Ventilation = Ventilation
        self.Loads = Loads
        self.Conditioning = Conditioning
        self.Constructions = Constructions
        self.DaylightMeshResolution = DaylightMeshResolution
        self.DaylightWorkplaneHeight = DaylightWorkplaneHeight
        self.DomesticHotWater = DomesticHotWater
        self.InternalMassConstruction = InternalMassConstruction
        self.InternalMassExposedPerFloorArea = InternalMassExposedPerFloorArea

        self.Windows = Windows  # This is not used in to_json()

        self._epbunch = kwargs.get("epbunch", None)
        self._zonesurfaces = kwargs.get("zonesurfaces", None)
        self._area = None
        self._volume = None
        self._occupants = None
        self._is_part_of_conditioned_floor_area = None
        self._is_part_of_total_floor_area = None
        self._multiplier = None

    @property
    def InternalMassExposedPerFloorArea(self):
        return float(self._InternalMassExposedPerFloorArea)

    @InternalMassExposedPerFloorArea.setter
    def InternalMassExposedPerFloorArea(self, value):
        self._InternalMassExposedPerFloorArea = value

    def __add__(self, other):
        """
        Args:
            other (ZoneDefinition):
        """
        # create the new merged zone from self
        return ZoneDefinition.combine(self, other)

    def __hash__(self):
        return hash(
            (self.__class__.__name__, getattr(self, "Name", None), self.DataSource)
        )

    def __eq__(self, other):
        if not isinstance(other, ZoneDefinition):
            return NotImplemented
        else:
            return all(
                [
                    self.Conditioning == other.Conditioning,
                    self.Constructions == other.Constructions,
                    self.DomesticHotWater == other.DomesticHotWater,
                    self.Loads == other.Loads,
                    self.Ventilation == other.Ventilation,
                    self.Windows == other.Windows,
                    self.InternalMassConstruction == other.InternalMassConstruction,
                    self.InternalMassExposedPerFloorArea
                    == other.InternalMassExposedPerFloorArea,
                    self.DaylightMeshResolution == other.DaylightMeshResolution,
                    self.DaylightWorkplaneHeight == other.DaylightWorkplaneHeight,
                ]
            )

    @property
    def occupants(self):
        if self._occupants is None:
            with sqlite3.connect(self.idf.sql_file) as conn:
                sql_query = (
                    "SELECT t.Value FROM TabularDataWithStrings t "
                    "WHERE TableName='Average Outdoor Air During Occupied Hours' and ColumnName='Nominal Number of Occupants' and RowName=?"
                )

                fetchone = conn.execute(sql_query, (self.Name.upper(),)).fetchone()
                (res,) = fetchone or (0,)
            self._occupants = float(res)
        return self._occupants

    @occupants.setter
    def occupants(self, value):
        self._occupants = value

    @property
    def area(self):
        if self._area is None:
            with sqlite3.connect(self.idf.sql_file) as conn:
                sql_query = """
                    SELECT t.Value 
                    FROM TabularDataWithStrings t 
                    WHERE TableName='Zone Summary' and ColumnName='Area' and RowName=?
                """
                (res,) = conn.execute(sql_query, (self.Name.upper(),)).fetchone()
            self._area = float(res)
        return self._area

    @area.setter
    def area(self, value):
        self._area = value

    @property
    def volume(self):
        """Calculates the volume of the zone

        Returns (float): zone's volume in m³
        """
        if self._volume is None:
            with sqlite3.connect(self.idf.sql_file) as conn:
                sql_query = (
                    "SELECT t.Value FROM TabularDataWithStrings t "
                    "WHERE TableName='Zone Summary' and ColumnName='Volume' and RowName=?"
                )
                (res,) = conn.execute(sql_query, (self.Name.upper(),)).fetchone()
            self._volume = float(res)
        return self._volume

    @volume.setter
    def volume(self, value):
        self._volume = value

    def zonesurfaces(self, exclude=None):
        """Returns list of surfaces belonging to this zone. Optionally filter
        surface types.

        Args:
            exclude (list): exclude surface types, e.g.: ["INTERNALMASS",
                "WINDOWSHADINGCONTROL"]. Object key must be in capital letters.
        """
        if exclude is None:
            exclude = []
        if self._zonesurfaces is None:
            self._zonesurfaces = [surf for surf in self._epbunch.zonesurfaces]
        return [surf for surf in self._zonesurfaces if surf.key.upper() not in exclude]

    @property
    def is_core(self):
        return is_core(self._epbunch)

    @property
    def multiplier(self):
        """Zone multipliers are designed as a “multiplier” for floor area,
        zone loads, and energy consumed by internal gains.
        """
        if self._multiplier is None:
            with sqlite3.connect(self.idf.sql_file) as conn:
                sql_query = "SELECT t.Value FROM TabularDataWithStrings t WHERE TableName='Zone Summary' and ColumnName='Multipliers' and RowName=?"
                (res,) = conn.execute(sql_query, (self.Name.upper(),)).fetchone()
            self._multiplier = int(float(res))
        return self._multiplier

    @multiplier.setter
    def multiplier(self, value):
        self._multiplier = value

    @property
    def is_part_of_conditioned_floor_area(self):
        """Returns True if zone is conditioned"""
        if self._is_part_of_conditioned_floor_area is None:
            with sqlite3.connect(self.idf.sql_file) as conn:
                sql_query = (
                    "SELECT t.Value FROM TabularDataWithStrings t WHERE TableName='Zone Summary' and ColumnName='Conditioned (Y/N)' and RowName=?"
                    ""
                )
                res = conn.execute(sql_query, (self.Name.upper(),)).fetchone()
            self._is_part_of_conditioned_floor_area = "Yes" in res
        return self._is_part_of_conditioned_floor_area

    @property
    def is_part_of_total_floor_area(self):
        """Returns True if zone is part of the total floor area"""
        if self._is_part_of_total_floor_area is None:
            with sqlite3.connect(self.idf.sql_file) as conn:
                sql_query = "SELECT t.Value FROM TabularDataWithStrings t WHERE TableName='Zone Summary' and ColumnName='Part of Total Floor Area (Y/N)' and RowName=?"
                res = conn.execute(sql_query, (self.Name.upper(),)).fetchone()
            self._is_part_of_total_floor_area = "Yes" in res
        return self._is_part_of_total_floor_area

    @staticmethod
    def get_volume_from_surfs(zone_surfs):
        """Calculate the volume of a zone only and only if the surfaces are such
        that you can find a point inside so that you can connect every vertex to
        the point without crossing a face.

        Adapted from: https://stackoverflow.com/a/19125446

        Args:
            zone_surfs (list): List of zone surfaces (EpBunch)
        """
        vol = 0
        for surf in zone_surfs:
            polygon_d = Polygon3D(surf.coords)  # create Polygon3D from surf
            n = len(polygon_d.vertices_list)
            v2 = polygon_d[0]
            x2 = v2.x
            y2 = v2.y
            z2 = v2.z

            for i in range(1, n - 1):
                v0 = polygon_d[i]
                x0 = v0.x
                y0 = v0.y
                z0 = v0.z
                v1 = polygon_d[i + 1]
                x1 = v1.x
                y1 = v1.y
                z1 = v1.z
                # Add volume of tetrahedron formed by triangle and origin
                vol += math.fabs(
                    x0 * y1 * z2
                    + x1 * y2 * z0
                    + x2 * y0 * z1
                    - x0 * y2 * z1
                    - x1 * y0 * z2
                    - x2 * y1 * z0
                )
        return vol / 6.0

    def to_json(self):
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Conditioning"] = self.Conditioning.to_dict()
        data_dict["Constructions"] = self.Constructions.to_dict()
        data_dict["DaylightMeshResolution"] = round(self.DaylightMeshResolution, 2)
        data_dict["DaylightWorkplaneHeight"] = round(self.DaylightWorkplaneHeight, 2)
        data_dict["DomesticHotWater"] = self.DomesticHotWater.to_dict()
        data_dict["InternalMassConstruction"] = self.InternalMassConstruction.to_dict()
        data_dict["InternalMassExposedPerFloorArea"] = round(
            self.InternalMassExposedPerFloorArea, 2
        )
        data_dict["Loads"] = self.Loads.to_dict()
        data_dict["Ventilation"] = self.Ventilation.to_dict()
        data_dict["Category"] = self.Category
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    @classmethod
    @deprecated(
        deprecated_in="1.3.1",
        removed_in="1.5",
        current_version=__version__,
        details="Use from_dict function instead",
    )
    def from_json(cls, *args, **kwargs):

        return cls.from_dict(*args, **kwargs)

    @classmethod
    def from_dict(
        cls,
        Conditioning,
        Constructions,
        DomesticHotWater,
        InternalMassConstruction,
        Loads,
        Ventilation,
        *args,
        **kwargs,
    ):
        """
        Args:
            *args:
            **kwargs:
        """
        Conditioning = cls.get_classref(Conditioning)
        Constructions = cls.get_classref(Constructions)
        DomesticHotWater = cls.get_classref(DomesticHotWater)
        InternalMassConstruction = cls.get_classref(InternalMassConstruction)
        Loads = cls.get_classref(Loads)
        Ventilation = cls.get_classref(Ventilation)
        zone = cls(
            *args,
            Conditioning=Conditioning,
            Constructions=Constructions,
            DomesticHotWater=DomesticHotWater,
            InternalMassConstruction=InternalMassConstruction,
            Loads=Loads,
            Ventilation=Ventilation,
            **kwargs,
        )

        return zone

    @classmethod
    def from_zone_epbunch(cls, zone_ep, construct_parents=True, **kwargs):
        """Create a Zone object from an eppy 'ZONE' epbunch.

        Args:
            zone_ep (eppy.bunch_subclass.EpBunch): The Zone EpBunch.
            construct_parents (bool): If False, skips construction of parents objects
                such as Constructions, Conditioning, etc.
        """
        start_time = time.time()
        log('Constructing :class:`Zone` for zone "{}"'.format(zone_ep.Name))
        name = zone_ep.Name
        zone = cls(
            Name=name,
            idf=zone_ep.theidf,
            Category=zone_ep.theidf.name,
            **kwargs,
        )

        zone._epbunch = zone_ep
        zone._zonesurfaces = zone_ep.zonesurfaces

        if construct_parents:
            zone.Constructions = ZoneConstructionSet.from_zone(zone, **kwargs)
            zone.Conditioning = ZoneConditioning.from_zone(zone, **kwargs)
            zone.Ventilation = VentilationSetting.from_zone(zone, **kwargs)
            zone.DomesticHotWater = DomesticHotWaterSetting.from_zone(zone, **kwargs)
            zone.Loads = ZoneLoad.from_zone(zone, **kwargs)
            zone.InternalMassConstruction = InternalMass.from_zone(zone, **kwargs)
            zone.Windows = WindowSetting.from_zone(zone, **kwargs)

        log(
            'completed Zone "{}" constructor in {:,.2f} seconds'.format(
                zone_ep.Name, time.time() - start_time
            )
        )
        return zone

    def combine(self, other, weights=None, allow_duplicates=False):
        """
        Args:
            other (ZoneDefinition):
            weights (list-like, optional): A list-like object of len 2. If None,
                the volume of the zones for which self and other belongs is
                used.

        Todo:
            Create Equivalent InternalMassConstruction from partitions when combining
            zones.

        Returns:
            (ZoneDefinition): the combined Zone object.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        meta = self._get_predecessors_meta(other)

        if not weights:
            zone_weight = settings.zone_weight
            weights = [
                getattr(self, str(zone_weight)),
                getattr(other, str(zone_weight)),
            ]
            log(
                'using zone {} "{}" as weighting factor in "{}" '
                "combine.".format(
                    zone_weight,
                    " & ".join(list(map(str, map(int, weights)))),
                    self.__class__.__name__,
                )
            )

        new_attr = dict(
            Conditioning=ZoneConditioning.combine(
                self.Conditioning, other.Conditioning, weights
            ),
            Constructions=ZoneConstructionSet.combine(
                self.Constructions, other.Constructions, weights
            ),
            Ventilation=VentilationSetting.combine(
                self.Ventilation, other.Ventilation, weights
            ),
            Windows=WindowSetting.combine(self.Windows, other.Windows, weights),
            DaylightMeshResolution=self._float_mean(
                other, "DaylightMeshResolution", weights=weights
            ),
            DaylightWorkplaneHeight=self._float_mean(
                other, "DaylightWorkplaneHeight", weights
            ),
            DomesticHotWater=DomesticHotWaterSetting.combine(
                self.DomesticHotWater, other.DomesticHotWater, weights
            ),
            InternalMassConstruction=OpaqueConstruction.combine(
                self.InternalMassConstruction, other.InternalMassConstruction
            ),
            InternalMassExposedPerFloorArea=self._float_mean(
                other, "InternalMassExposedPerFloorArea", weights
            ),
            Loads=ZoneLoad.combine(self.Loads, other.Loads, weights),
        )
        new_obj = ZoneDefinition(**meta, **new_attr, idf=self.idf)

        # transfer aggregated values [volume, area, occupants] to new combined zone
        new_obj.volume = self.volume + other.volume
        new_obj.area = self.area + other.area
        new_obj.occupants = self.occupants + other.occupants

        if new_attr["Conditioning"]:  # Could be None
            new_attr["Conditioning"]._belongs_to_zone = new_obj
        if new_attr["Constructions"]:  # Could be None
            new_attr["Constructions"]._belongs_to_zone = new_obj
        if new_attr["Ventilation"]:  # Could be None
            new_attr["Ventilation"]._belongs_to_zone = new_obj
        if new_attr["DomesticHotWater"]:  # Could be None
            new_attr["DomesticHotWater"]._belongs_to_zone = new_obj
        if new_attr["Windows"]:  # Could be None
            new_attr["Windows"]._belongs_to_zone = new_obj

        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def validate(self):
        """Validate object and fill in missing values."""
        if not self.InternalMassConstruction:
            InternalMass.generic_internalmass_from_zone(self)
            log(
                f"While validating {self}, the required attribute "
                f"'InternalMassConstruction' was filled "
                f"with {self.InternalMassConstruction} and the "
                f"'InternalMassExposedPerFloorArea' set to"
                f" {self.InternalMassExposedPerFloorArea}"
            )

        if self.Conditioning is None:
            self.Conditioning = ZoneConditioning(Name="Unconditioned Zone")

        return self

    def mapping(self):
        self.validate()

        return dict(
            Conditioning=self.Conditioning,
            Constructions=self.Constructions,
            DaylightMeshResolution=self.DaylightMeshResolution,
            DaylightWorkplaneHeight=self.DaylightWorkplaneHeight,
            DomesticHotWater=self.DomesticHotWater,
            InternalMassConstruction=self.InternalMassConstruction,
            InternalMassExposedPerFloorArea=self.InternalMassExposedPerFloorArea,
            Windows=self.Windows,
            Loads=self.Loads,
            Ventilation=self.Ventilation,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def get_ref(self, ref):
        """Get item matching reference id.

        Args:
            ref:
        """
        return next(
            iter(
                [
                    value
                    for value in ZoneDefinition.CREATED_OBJECTS
                    if value.id == ref["$ref"]
                ]
            ),
            None,
        )


def resolve_obco(this):
    """Resolve the outside boundary condition of a surface and return the other
    SURFACE epbunch and, if possible, the ZONE epbunch.

    Args:
        this (EpBunch): The surface for which we are identifying the boundary
            object.

    Returns:
        (EpBunch, EpBunch): A tuple of:

            EpBunch: The other surface EpBunch: The other zone

    Notes:
        Info on the Outside Boundary Condition Object of a surface of type
        BuildingSurface:Detailed:

        Non-blank only if the field `Outside Boundary Condition` is *Surface*,
        *Zone*, *OtherSideCoefficients* or *OtherSideConditionsModel*. If
        Surface, specify name of corresponding surface in adjacent zone or
        specify current surface name for internal partition separating like
        zones. If Zone, specify the name of the corresponding zone and the
        program will generate the corresponding interzone surface. If
        Foundation, specify the name of the corresponding Foundation object and
        the program will calculate the heat transfer appropriately. If
        OtherSideCoefficients, specify name of
        SurfaceProperty:OtherSideCoefficients. If OtherSideConditionsModel,
        specify name of SurfaceProperty:OtherSideConditionsModel.
    """

    # other belongs to which zone?
    # for key in this.getfieldidd_item('Outside_Boundary_Condition_Object',
    #                                  'validobjects'):

    obc = this.Outside_Boundary_Condition

    if obc.upper() == "ZONE":
        name = this.Outside_Boundary_Condition_Object
        adj_zone = this.theidf.getobject("ZONE", name)
        return None, adj_zone

    elif obc.upper() == "SURFACE":
        obco = this.get_referenced_object("Outside_Boundary_Condition_Object")
        adj_zone = obco.theidf.getobject("ZONE", obco.Zone_Name)
        return obco, adj_zone
    else:
        return None, None


def label_surface(row):
    """Takes a boundary and returns its corresponding umi-Category

    Args:
        row:
    """
    # Floors
    if row["Surface_Type"] == "Floor":
        if row["Outside_Boundary_Condition"] == "Surface":
            return "Interior Floor"
        if row["Outside_Boundary_Condition"] == "Ground":
            return "Ground Floor"
        if row["Outside_Boundary_Condition"] == "Outdoors":
            return "Exterior Floor"
        if row["Outside_Boundary_Condition"] == "Adiabatic":
            return "Interior Floor"
        else:
            return "Other"

    # Roofs & Ceilings
    if row["Surface_Type"] == "Roof":
        return "Roof"
    if row["Surface_Type"] == "Ceiling":
        return "Interior Floor"
    # Walls
    if row["Surface_Type"] == "Wall":
        if row["Outside_Boundary_Condition"] == "Surface":
            return "Partition"
        if row["Outside_Boundary_Condition"] == "Outdoors":
            return "Facade"
        if row["Outside_Boundary_Condition"] == "Adiabatic":
            return "Partition"
    return "Other"


def type_surface(row):
    """Takes a boundary and returns its corresponding umi-type

    Args:
        row:
    """

    # Floors
    if row["Surface_Type"] == "Floor":
        if row["Outside_Boundary_Condition"] == "Surface":
            return 3  # umi defined
        if row["Outside_Boundary_Condition"] == "Ground":
            return 2  # umi defined
        if row["Outside_Boundary_Condition"] == "Outdoors":
            return 4  # umi defined
        if row["Outside_Boundary_Condition"] == "Adiabatic":
            return 5
        else:
            return ValueError('Cannot find Construction Type for "{}"'.format(row))

    # Roofs & Ceilings
    elif row["Surface_Type"] == "Roof":
        return 1
    elif row["Surface_Type"] == "Ceiling":
        return 3
    # Walls
    elif row["Surface_Type"] == "Wall":
        if row["Outside_Boundary_Condition"] == "Surface":
            return 5  # umi defined
        if row["Outside_Boundary_Condition"] == "Outdoors":
            return 0  # umi defined
        if row["Outside_Boundary_Condition"] == "Adiabatic":
            return 5  # umi defined
    else:
        raise ValueError('Cannot find Construction Type for "{}"'.format(row))


def zone_information(df):
    """Each zone_loads is summarized in a simple set of statements

    Args:
        df:

    Returns:
        df

    References:
        * ` Zone Loads Information

        < https://bigladdersoftware.com/epx/docs/8-3/output-details-and
        -examples/eplusout.eio.html#zone_loads-information>`_
    """
    df = get_from_tabulardata(df)
    tbstr = df[
        (df.ReportName == "Initialization Summary")
        & (df.TableName == "Zone Information")
    ].reset_index()
    # Ignore Zone that are not part of building area
    pivoted = tbstr.pivot_table(
        index=["RowName"],
        columns="ColumnName",
        values="Value",
        aggfunc=lambda x: " ".join(x),
    )

    return pivoted.loc[pivoted["Part of Total Building Area"] == "Yes", :]


def get_from_tabulardata(sql):
    """Returns a DataFrame from the 'TabularDataWithStrings' table.

    Args:
        sql (dict):

    Returns:
        (pandas.DataFrame)
    """
    tab_data_wstring = sql["TabularDataWithStrings"]
    tab_data_wstring.index.names = ["Index"]

    # strip whitespaces
    tab_data_wstring.Value = tab_data_wstring.Value.str.strip()
    tab_data_wstring.RowName = tab_data_wstring.RowName.str.strip()
    return tab_data_wstring


def is_core(zone):
    """

    Args:
        zone (eppy.bunch_subclass.EpBunch): The Zone object.

    Returns:
        (bool): Whether the zone is a core zone or not.
    """
    # if all surfaces don't have boundary condition == "Outdoors"
    iscore = True
    for s in zone.zonesurfaces:
        try:
            if (abs(int(s.tilt)) < 180) & (abs(int(s.tilt)) > 0):
                obc = s.Outside_Boundary_Condition.lower()
                if obc in ["outdoors", "ground"]:
                    iscore = False
                    break
        except BadEPFieldError:
            pass  # pass surfaces that don't have an OBC,
            # eg. InternalMass
    return iscore


def iscore(row):
    """Helps to group by core and perimeter zones. If any of "has `core` in
    name" and "ExtGrossWallArea == 0" is true, will consider zone_loads as core,
    else as perimeter.

    Todo:
        * assumes a basement zone_loads will be considered as a core zone_loads
          since no ext wall area for basements.

    Args:
        row (pandas.Series): a row

    Returns:
        str: 'Core' or 'Perimeter'
    """
    if any(
        [
            "core" in row["Zone Name"].lower(),
            float(row["Exterior Gross Wall Area {m2}"]) == 0,
        ]
    ):
        # We look for the string `core` in the Zone_Name
        return "Core"
    elif row["Part of Total Building Area"] == "No":
        return np.NaN
    elif "plenum" in row["Zone Name"].lower():
        return np.NaN
    else:
        return "Perimeter"



################################################################################
# Module: archetypal.template
# Description:
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import collections
import hashlib

import numpy as np
import pandas as pd
from deprecation import deprecated
from eppy.bunch_subclass import EpBunch

import archetypal
from archetypal import Schedule, log
from archetypal.template import UmiBase, UniqueName


class UmiSchedule(Schedule, UmiBase):
    """Class that handles Schedules as"""

    def __init__(self, *args, quantity=None, **kwargs):
        """
        Args:
            *args:
            **kwargs:
        """
        super(UmiSchedule, self).__init__(**kwargs)
        self.quantity = quantity

    @classmethod
    def constant_schedule(
        cls, hourly_value=1, Name="AlwaysOn", Type="Fraction", idf=None, **kwargs
    ):
        """
        Args:
            hourly_value:
            Name:
            idf:
            **kwargs:
        """
        return super(UmiSchedule, cls).constant_schedule(
            hourly_value=hourly_value, Name=Name, Type=Type, idf=idf, **kwargs
        )

    @classmethod
    def from_values(cls, Name, Values, idf, Type="Fraction", **kwargs):
        """
        Args:
            Name:
            Values:
            idf:
            Type:
            **kwargs:
        """
        return super(UmiSchedule, cls).from_values(
            Name=Name, Values=Values, Type=Type, idf=idf, **kwargs
        )

    @classmethod
    def from_yearschedule(cls, year_sched, idf=None):
        """
        Args:
            year_sched:
            idf:
        """
        if isinstance(year_sched, YearSchedule):
            return cls.from_values(
                Name=year_sched.Name,
                Values=year_sched.all_values,
                Type=year_sched.Type,
                idf=idf,
            )

    def __add__(self, other):
        return UmiSchedule.combine(self, other)

    def __repr__(self):
        name = self.Name
        resample = self.series.resample("D")
        min = resample.min().mean()
        mean = resample.mean().mean()
        max = resample.max().mean()
        return (
            name
            + ": "
            + "mean daily min:{:.2f} mean:{:.2f} max:{:.2f} ".format(min, mean, max)
            + (f"quantity {self.quantity}" if self.quantity is not None else "")
        )

    def __str__(self):
        return repr(self)

    def __hash__(self):
        return hash((self.__class__.__name__, getattr(self, "Name", None)))

    def __eq__(self, other):
        if not isinstance(other, UmiSchedule):
            return NotImplemented
        else:
            return all(
                [
                    # self.Name == other.Name,
                    self.strict == other.strict,
                    self.schType == other.schType,
                    self.Type == other.Type,
                    self.quantity == other.quantity,
                    np.allclose(self.all_values, other.all_values, rtol=1e-02)
                    if self.all_values.size == other.all_values.size
                    else False,
                ]
            )

    def combine(self, other, weights=None, quantity=None):
        """Combine two UmiSchedule objects together.

        Args:
            other (UmiSchedule): The other Schedule object to combine with.
            weights (list, dict or string): Attribute of self and other containing the
                weight factor. If a list is passed, it must have len = 2; the first
                element is applied to self and the second element is applied to other.
                If a dict is passed, the self.Name and other.Name are the keys. If a
                str is passed, the
            quantity (list or dict or bool): Scalar value that will be multiplied by
                self before the averaging occurs. This ensures that the resulting
                schedule returns the correct integrated value. If a dict is passed,
                keys are schedules Names and values are quantities.

        Returns:
            (UmiSchedule): the combined UmiSchedule object.

        Raises:
            TypeError: if Quantity is not of type list, tuple, dict or a callable.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        if not isinstance(other, UmiSchedule):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # check if the schedule is the same
        if self == other:
            if self.quantity and other.quantity:
                self.quantity += other.quantity
            return self
        # check if self is only zeros. Should not affect other.
        if not np.any(self.all_values):
            return other
        # check if other is only zeros. Should not affect self.
        if not np.any(other.all_values):
            return self

        if not weights:
            log(
                'using 1 as weighting factor in "{}" '
                "combine.".format(self.__class__.__name__)
            )
            weights = [1, 1]
        elif isinstance(weights, str):
            # get the attribute from self and other
            weights = [getattr(self, weights), getattr(other, weights)]
        elif isinstance(weights, (list, tuple)):
            # check if length is 2.
            l = len(weights)
            if l != 2:
                raise ValueError(
                    "USing a list or tuple, the weights attribute must "
                    "have a length of 2. A length of {}".format(l)
                )
        elif isinstance(weights, dict):
            weights = [weights[self.Name], weights[other.Name]]

        if quantity is None:
            new_values = np.average(
                [self.all_values, other.all_values], axis=0, weights=weights
            )
        elif isinstance(quantity, dict):
            # Multiplying the schedule values by the quantity for both self and other
            # and then using a weighted average. Finally, new values are normalized.
            new_values = np.average(
                [
                    self.all_values * quantity[self.Name],
                    other.all_values * quantity[other.Name],
                ],
                axis=0,
                weights=weights,
            )
            new_values /= quantity[self.Name] + quantity[other.Name]
        elif callable(quantity):
            new_values = np.average(
                np.stack((self.all_values, other.all_values), axis=1),
                axis=1,
                weights=[
                    quantity(self.predecessors.data),
                    quantity(other.predecessors.data),
                ],
            )
        elif isinstance(quantity, (list, tuple)):
            # Multiplying the schedule values by the quantity for both self and other
            # and then using a weighted average. Finally, new values are normalized.
            self_quantity, other_quantity = quantity
            new_values = (
                self.all_values * self_quantity + other.all_values * other_quantity
            ) / sum(quantity)
        elif isinstance(quantity, bool):
            self_quantity, other_quantity = self.quantity, other.quantity
            new_values = (
                self.all_values * self_quantity + other.all_values * other_quantity
            ) / (self_quantity + other_quantity)
        else:
            raise TypeError("Quantity is not of type list, tuple, dict or a callable")

        # the new object's name
        meta = self._get_predecessors_meta(other)

        # Overriding meta Name
        hasher = hashlib.md5()
        hasher.update(new_values)
        meta["Name"] = f"Combined_UmiSchedule_{hasher.hexdigest()}"
        quantity = np.nansum(
            [self.quantity or float("nan"), other.quantity or float("nan")]
        )
        new_obj = UmiSchedule.from_values(
            Values=new_values, Type="Fraction", quantity=quantity, idf=self.idf, **meta
        )
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        new_obj.weights = sum(weights)
        return new_obj

    def develop(self):
        year, weeks, days = self.to_year_week_day()
        lines = ["- {}".format(obj) for obj in self.predecessors]

        newdays = []
        for day in days:
            newdays.append(
                DaySchedule.from_epbunch(
                    day,
                    Comments="Year Week Day schedules created from: \n{}".format(
                        "\n".join(lines)
                    ),
                    allow_duplicates=getattr(self, "_not_unique", False),
                    Category=self.Name,
                )
            )
        Parts = []
        weeks = {schd.Name: schd for schd in weeks}

        def chunks(lst, n):
            """Yield successive n-sized chunks from lst."""
            for i in range(0, len(lst), n):
                yield lst[i : i + n]

        for fields in chunks(year.fieldvalues[3:], 5):
            weekname, from_month, from_day, to_month, to_day = fields
            Parts.append(
                YearSchedulePart(
                    FromMonth=from_month,
                    ToMonth=to_month,
                    FromDay=from_day,
                    ToDay=to_day,
                    Schedule=WeekSchedule.from_epbunch(
                        weeks[weekname],
                        Comments="Year Week Day schedules created from:\n{}".format(
                            "\n".join(lines)
                        ),
                        Category=self.Name,
                        allow_duplicates=getattr(self, "_not_unique", False),
                    ),
                )
            )
        _from = "\n".join(lines)
        return YearSchedule(
            Name=self.Name,
            Parts=Parts,
            Type="Fraction",
            epbunch=year,
            Category=self.Name,
            Comments=f"Year Week Day schedules created from: \n{_from}" + str(id(self)),
            idf=self.idf,
            allow_duplicates=getattr(self, "_not_unique", False),
        )

    def get_unique(self):
        return super(UmiSchedule, self.develop()).get_unique()

    def to_json(self):
        """UmiSchedule does not implement the to_json method because it is not
        used when generating the json file. Only Year-Week- and DaySchedule
        classes are used
        """

        return self.to_dict()

    def to_dict(self):
        return {"$ref": str(self.id)}

    def validate(self):
        """Validate object and fill in missing values."""
        return self

    def mapping(self):
        self.validate()

        return dict(
            Category=self.Category,
            Type=self.Type,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def get_ref(self, ref):
        """Get item matching reference id.

        Args:
            ref:
        """
        return next(
            iter(
                [
                    value
                    for value in UmiSchedule.CREATED_OBJECTS
                    if value.id == ref["$ref"]
                ]
            ),
            None,
        )


class YearSchedulePart:
    """Helper Class for YearSchedules that are defined using FromDay FromMonth
    ToDay ToMonth attributes.
    """

    def __init__(
        self,
        FromDay=None,
        FromMonth=None,
        ToDay=None,
        ToMonth=None,
        Schedule=None,
        **kwargs,
    ):
        """
        Args:
            FromDay (int): This numeric field is the starting day for the
                schedule time period.
            FromMonth (int): This numeric field is the starting month for the
                schedule time period.
            ToDay (int): This numeric field is the ending day for the schedule
                time period.
            ToMonth (int): This numeric field is the ending month for the
                schedule time period.
            Schedule (UmiSchedule): The associated UmiSchedule related to this
                object.
            kwargs (dict): Other Keyword arguments.
        """
        self.FromDay = FromDay
        self.FromMonth = FromMonth
        self.ToDay = ToDay
        self.ToMonth = ToMonth
        self.Schedule = Schedule

    def __eq__(self, other):
        if not isinstance(other, YearSchedulePart):
            return NotImplemented
        else:
            return all(
                [
                    self.FromDay == other.FromDay,
                    self.FromMonth == other.FromMonth,
                    self.ToDay == other.ToDay,
                    self.ToMonth == other.ToMonth,
                    self.Schedule == other.Schedule,
                ]
            )

    def __iter__(self):
        for k, v in self.mapping().items():
            yield k, v

    def __hash__(self):
        return id(self)

    @classmethod
    @deprecated(
        deprecated_in="1.3.1",
        removed_in="1.5",
        current_version=archetypal.__version__,
        details="Use from_dict function instead",
    )
    def from_json(cls, *args, **kwargs):

        """
        Args:
            all_objects:
            *args:
            **kwargs:
        """
        return cls.from_dict(*args, **kwargs)

    @classmethod
    def from_dict(cls, Schedule, **kwargs):
        """
        Args:
            all_objects:
            *args:
            **kwargs:
        """
        ref = UmiBase.get_classref(Schedule)
        ysp = cls(Schedule=ref, **kwargs)

        return ysp

    def to_dict(self):
        return collections.OrderedDict(
            FromDay=self.FromDay,
            FromMonth=self.FromMonth,
            ToDay=self.ToDay,
            ToMonth=self.ToMonth,
            Schedule={"$ref": str(self.Schedule.id)},
        )

    def __str__(self):
        return str(self.to_dict())

    def mapping(self):
        return dict(
            FromDay=self.FromDay,
            FromMonth=self.FromMonth,
            ToDay=self.ToDay,
            ToMonth=self.ToMonth,
            Schedule=self.Schedule,
        )

    def get_unique(self):
        return self


class DaySchedule(UmiSchedule):
    """Superclass of UmiSchedule that handles daily schedules."""

    def __init__(self, Category="Day", **kwargs):
        """Initialize a DaySchedule object with parameters:

        Args:
            Category:
            **kwargs: Keywords passed to the :class:`UmiSchedule` constructor.
        """
        super(DaySchedule, self).__init__(Category=Category, **kwargs)

    def __eq__(self, other):
        if not isinstance(other, DaySchedule):
            return NotImplemented
        else:
            return all(
                [
                    self.Type == other.Type,
                    np.allclose(self.all_values, other.all_values, rtol=1e-02),
                ]
            )

    def __hash__(self):
        return super(DaySchedule, self).__hash__()

    @classmethod
    def from_epbunch(cls, epbunch, **kwargs):
        """Create a DaySchedule from a :class:`~eppy.bunch_subclass.EpBunch`
        object

        Args:
            epbunch (EpBunch): The EpBunch object to construct a DaySchedule
                from.
            **kwargs: Keywords passed to the :class:`UmiSchedule` constructor.
                See :class:`UmiSchedule` for more details.
        """

        sched = cls(
            idf=epbunch.theidf,
            Name=epbunch.Name,
            epbunch=epbunch,
            schType=epbunch.key,
            **kwargs,
        )

        return sched

    @classmethod
    def from_values(cls, Name, Values, idf, Type="Fraction", **kwargs):
        """Create a DaySchedule from an array of size (24,)

        Args:
            Name:
            Values (array-like): A list of values of length 24.
            idf (IDF): The idf model.
            Type:
            **kwargs: Keywords passed to the :class:`UmiSchedule` constructor.
                See :class:`UmiSchedule` for more details.
        """
        return cls(Name=Name, Values=np.array(Values), Type=Type, idf=idf, **kwargs)

    @classmethod
    @deprecated(
        deprecated_in="1.3.1",
        removed_in="1.5",
        current_version=archetypal.__version__,
        details="Use from_dict function instead",
    )
    def from_json(cls, Type, **kwargs):

        """
        Args:
            Type:
            **kwargs:
        """
        return cls.from_dict(Type, **kwargs)

    @classmethod
    def from_dict(cls, Name, Values, Type, **kwargs):
        """Create a DaySchedule from a Umi Template json file.

        Args:
            Type (str): The schedule type limits name.
            **kwargs:
        """
        sched = cls.from_values(Name=Name, Values=Values, Type=Type, **kwargs)

        return sched

    def get_unique(self):
        return UmiBase.get_unique(self)

    def to_json(self):
        """Returns a dict-like representation of the schedule.

        Returns:
            dict: The dict-like representation of the schedule
        """

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Category"] = self.Category
        data_dict["Type"] = self.Type
        data_dict["Values"] = self.all_values.round(3).tolist()
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    def mapping(self):
        return dict(
            Category=self.Category,
            Type=self.Type,
            Values=self.all_values.round(3).tolist(),
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    @property
    def all_values(self) -> np.ndarray:
        if self._values is None:
            self._values = self.get_schedule_values(self.epbunch)
        if isinstance(self._values, list):
            self._values = np.array(self._values)
        return self._values

    def to_dict(self):
        """returns umi template repr"""
        return {"$ref": str(self.id)}


class WeekSchedule(UmiSchedule):
    """Superclass of UmiSchedule that handles weekly schedules."""

    def __init__(self, Days=None, Category="Week", **kwargs):
        """Initialize a WeekSchedule object with parameters:

        Args:
            Days (list of DaySchedule): list of :class:`DaySchedule`.
            **kwargs:
        """
        super(WeekSchedule, self).__init__(Category=Category, **kwargs)
        self.Days = Days

    def __eq__(self, other):
        if not isinstance(other, WeekSchedule):
            return NotImplemented
        else:
            return all(
                [
                    self.Type == other.Type,
                    self.Days == other.Days,
                ]
            )

    def __hash__(self):
        return super(WeekSchedule, self).__hash__()

    @classmethod
    def from_epbunch(cls, epbunch, **kwargs):
        """
        Args:
            epbunch:
            **kwargs:
        """
        Days = WeekSchedule.get_days(epbunch)
        sched = cls(
            idf=epbunch.theidf,
            Name=epbunch.Name,
            schType=epbunch.key,
            Days=Days,
            **kwargs,
        )

        return sched

    @classmethod
    @deprecated(
        deprecated_in="1.3.1",
        removed_in="1.5",
        current_version=archetypal.__version__,
        details="Use from_dict function instead",
    )
    def from_json(cls, **kwargs):

        """
        Args:
            **kwargs:
        """
        return cls.from_dict(**kwargs)

    @classmethod
    def from_dict(cls, Type, **kwargs):
        """
        Args:
            **kwargs:
        """
        refs = kwargs.pop("Days")
        Days = [UmiBase.get_classref(ref) for ref in refs]
        wc = cls(Type=Type, Days=Days, **kwargs)
        return wc

    def get_unique(self):
        return UmiBase.get_unique(self)

    def to_json(self):
        """Returns a dict-like representation of the schedule.

        Returns:
            dict: The dict-like representation of the schedule
        """
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Category"] = self.Category
        data_dict["Days"] = [day.to_dict() for day in self.Days]
        data_dict["Type"] = self.Type
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    def mapping(self):
        return dict(
            Category=self.Category,
            Days=self.Days,
            Type=self.Type,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    @classmethod
    def get_days(cls, epbunch):
        """
        Args:
            epbunch (EpBunch):
        """
        blocks = []
        dayname = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]
        for day in dayname:
            week_day_schedule_name = epbunch["{}_ScheduleDay_Name".format(day)]
            blocks.append(
                next(
                    (
                        x
                        for x in UmiBase.CREATED_OBJECTS
                        if x.Name == week_day_schedule_name
                        and type(x).__name__ == "DaySchedule"
                    ),
                    None,
                )
            )

        return blocks

    @property
    def all_values(self) -> np.ndarray:
        if self._values is None:
            self._values = np.concatenate([day.all_values for day in self.Days])
        return self._values

    def to_dict(self):
        """returns umi template repr"""
        return {"$ref": str(self.id)}


class YearSchedule(UmiSchedule):
    """Superclass of UmiSchedule that handles yearly schedules."""

    def __init__(self, Name, Type="Fraction", Parts=None, Category="Year", **kwargs):
        """Initialize a YearSchedule object with parameters:

        Args:
            Category:
            Name:
            Type:
            Parts (list of YearSchedulePart): The YearScheduleParts.
            **kwargs:
        """
        self.epbunch = kwargs.get("epbunch", None)
        if Parts is None:
            self.Parts = self.get_parts(self.epbunch)
        else:
            self.Parts = Parts
        super(YearSchedule, self).__init__(
            Name=Name, Type=Type, schType="Schedule:Year", Category=Category, **kwargs
        )

    def __eq__(self, other):
        if not isinstance(other, YearSchedule):
            return NotImplemented
        else:
            return all([self.Type == other.Type, self.Parts == other.Parts])

    def __hash__(self):
        return super(YearSchedule, self).__hash__()

    @classmethod
    def from_parts(cls, *args, Parts, **kwargs):
        """
        Args:
            *args:
            Parts (list of YearSchedulePart):
            **kwargs:
        """
        ysp = cls(*args, Parts=Parts, **kwargs)
        ysp._values = ysp.all_values

        return ysp

    @property
    def all_values(self) -> np.ndarray:
        if self._values is None:
            index = pd.date_range(start=self.startDate, freq="1H", periods=8760)
            series = pd.Series(index=index)
            for part in self.Parts:
                start = "{}-{}-{}".format(self.year, part.FromMonth, part.FromDay)
                end = "{}-{}-{}".format(self.year, part.ToMonth, part.ToDay)
                try:  # Get week values from all_values of Days that are DaySchedule object
                    one_week = np.array(
                        [
                            item
                            for sublist in part.Schedule.Days
                            for item in sublist.all_values
                        ]
                    )
                except:  # Days are not DaySchedule object
                    try:  # Days is a list of 7 dicts (7 days in a week)
                        # Dicts are the id of Days ({"$ref": id})
                        day_values = [self.get_ref(day) for day in part.Schedule.Days]
                        values = []
                        for i in range(0, 7):  # There is 7 days a week
                            values = values + day_values[i].all_values.tolist()
                        one_week = np.array(values)
                    except:
                        msg = (
                            'Days are not a DaySchedule or dictionaries in the form "{'
                            '$ref: id}" '
                        )
                        raise NotImplementedError(msg)

                all_weeks = np.resize(one_week, len(series.loc[start:end]))
                series.loc[start:end] = all_weeks
            self._values = series.values
        return self._values

    @classmethod
    @deprecated(
        deprecated_in="1.3.1",
        removed_in="1.5",
        current_version=archetypal.__version__,
        details="Use from_dict function instead",
    )
    def from_json(cls, **kwargs):

        """
        Args:
            **kwargs:
        """
        return cls.from_dict(**kwargs)

    @classmethod
    def from_dict(cls, Type, **kwargs):
        """
        Args:
            **kwargs:
        """
        Parts = [
            YearSchedulePart.from_dict(**part) for part in kwargs.pop("Parts", None)
        ]
        ys = cls(Type=Type, Parts=Parts, **kwargs)
        ys.schType = "Schedule:Year"
        return ys

    def get_unique(self):
        return UmiBase.get_unique(self)

    def to_json(self):
        """Returns a dict-like representation of the schedule.

        Returns:
            dict: The dict-like representation of the schedule
        """
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Category"] = self.Category
        data_dict["Parts"] = [part.to_dict() for part in self.Parts]
        data_dict["Type"] = self.Type
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    def mapping(self):
        self.validate()

        return dict(
            Category=self.Category,
            Parts=self.Parts,
            Type=self.Type,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def get_parts(self, epbunch):
        """
        Args:
            epbunch (EpBunch):
        """
        parts = []
        for i in range(int(len(epbunch.fieldvalues[3:]) / 5)):
            week_day_schedule_name = epbunch["ScheduleWeek_Name_{}".format(i + 1)]

            FromMonth = epbunch["Start_Month_{}".format(i + 1)]
            ToMonth = epbunch["End_Month_{}".format(i + 1)]
            FromDay = epbunch["Start_Day_{}".format(i + 1)]
            ToDay = epbunch["End_Day_{}".format(i + 1)]
            parts.append(
                YearSchedulePart(
                    FromDay,
                    FromMonth,
                    ToDay,
                    ToMonth,
                    next(
                        (
                            x
                            for x in self.CREATED_OBJECTS
                            if x.Name == week_day_schedule_name
                            and type(x).__name__ == "WeekSchedule"
                        )
                    ),
                )
            )
        return parts

    def to_dict(self):
        """returns umi template repr"""
        return {"$ref": str(self.id)}
